from typing import Annotated, Literal

from langchain_anthropic import ChatAnthropic
from langchain_tavily import TavilySearch
from langchain_core.messages import HumanMessage, SystemMessage, BaseMessage
from langchain_core.tools import tool
from langchain_experimental.utilities import PythonREPL
from langgraph.graph import END, START, MessagesState, StateGraph
from langgraph.prebuilt import create_react_agent
from langgraph.types import Command
from pydantic import BaseModel
from typing_extensions import TypedDict


tavily_tool = TavilySearch(max_results=5)

# This executes code locally, which can be unsafe
repl = PythonREPL()


@tool
def python_repl_tool(
    code: Annotated[str, "The python code to execute to generate your chart."],
):
    """Use this to execute python code and do math. If you want to see the output of a value,
    you should print it out with `print(...)`. This is visible to the user.
    """
    try:
        result = repl.run(code)
    except BaseException as e:
        return f"Failed to execute. Error: {repr(e)}"
    result_str = f"Successfully executed:\n```python\n{code}\n```\nStdout: {result}"
    return result_str

members = ["researcher", "coder"]
# Our team supervisor is an LLM node. It just picks the next agent to process
# and decides when the work is completed
options = members + ["FINISH"]

system_prompt = (
    "You are a supervisor tasked with managing a conversation between the"
    f" following workers: {members}. Given the following user request,"
    " respond with the worker to act next. Each worker will perform a"
    " task and respond with their results and status. When finished,"
    " respond with FINISH."
)


class Router(TypedDict):
    """Worker to route to next. If no workers needed, route to FINISH."""

    next: Literal[*options]


llm = ChatAnthropic(model="claude-3-5-sonnet-latest")

class GraphInput(BaseModel):
    question: str

class GraphOutput(BaseModel):
    answer: str

class State(MessagesState):
    next: str

def get_message_text(msg: BaseMessage) -> str:
    """LangChain-style safe message text extractor."""
    if isinstance(msg.content, str):
        return msg.content
    if isinstance(msg.content, list):
        return "".join(
            block.get("text", "") for block in msg.content if block.get("type") == "text"
        )
    return ""

def input(state: GraphInput):
    return {
        "messages": [
            SystemMessage(content=system_prompt),
            HumanMessage(content=state.question),
        ],
        "next": "",
    }

async def supervisor_node(state: State) -> Command[Literal[*members]] | GraphOutput:
    response = await llm.with_structured_output(Router).ainvoke(state["messages"])
    goto = response["next"]
    if goto == "FINISH":
        return GraphOutput(answer=get_message_text(state["messages"][-1]))
    else:
        return Command(goto=goto, update={"next": goto})

research_agent = create_react_agent(
    llm, tools=[tavily_tool], prompt="You are a researcher. DO NOT do any math."
)


async def research_node(state: State) -> Command[Literal["supervisor"]]:
    result = await research_agent.ainvoke(state)
    return Command(
        update={
            "messages": [
                HumanMessage(content=result["messages"][-1].content, name="researcher")
            ]
        },
        goto="supervisor",
    )


# NOTE: THIS PERFORMS ARBITRARY CODE EXECUTION, WHICH CAN BE UNSAFE WHEN NOT SANDBOXED
code_agent = create_react_agent(llm, tools=[python_repl_tool])


async def code_node(state: State) -> Command[Literal["supervisor"]]:
    result = await code_agent.ainvoke(state)
    return Command(
        update={
            "messages": [
                HumanMessage(content=result["messages"][-1].content, name="coder")
            ]
        },
        goto="supervisor",
    )


builder = StateGraph(State, input=GraphInput, output=GraphOutput)
builder.add_edge(START, "input")
builder.add_edge("input", "supervisor")
builder.add_node("input", input)
builder.add_node("supervisor", supervisor_node)
builder.add_node("researcher", research_node)
builder.add_node("coder", code_node)
graph = builder.compile()



"""Tool node factory wiring directly to LangGraph's ToolNode."""

from collections.abc import Sequence

from langchain_core.tools import BaseTool
from langgraph.prebuilt import ToolNode


def create_tool_node(tools: Sequence[BaseTool]) -> dict[str, ToolNode]:
    """Create individual ToolNode for each tool.

    Args:
        tools: Sequence of tools to create nodes for.

    Returns:
        Dict mapping tool.name -> ToolNode([tool]).
        Each tool gets its own dedicated node for middleware composition.

    Note:
        handle_tool_errors=False delegates error handling to LangGraph's error boundary.
    """
    return {tool.name: ToolNode([tool], handle_tool_errors=False) for tool in tools}



import os
from typing import Sequence, Type

from langchain_core.language_models import BaseChatModel
from langchain_core.messages import HumanMessage, SystemMessage
from langchain_core.tools import BaseTool
from langgraph.constants import END, START
from langgraph.graph import StateGraph
from pydantic import BaseModel

from ..tools import create_tool_node
from .init_node import (
    create_init_node,
)
from .llm_node import (
    create_llm_node,
)
from .router import (
    route_agent,
)
from .terminate_node import (
    create_terminate_node,
)
from .tools import create_flow_control_tools
from .types import AgentGraphConfig, AgentGraphNode, AgentGraphState


def create_agent(
    model: BaseChatModel,
    tools: Sequence[BaseTool],
    messages: Sequence[SystemMessage | HumanMessage],
    *,
    state_schema: Type[AgentGraphState] = AgentGraphState,
    response_format: type[BaseModel] | None = None,
    config: AgentGraphConfig | None = None,
) -> StateGraph[AgentGraphState]:
    """Build agent graph with INIT -> AGENT <-> TOOLS loop, terminated by control flow tools.

    Control flow tools (end_execution, raise_error) are auto-injected alongside regular tools.
    """
    if config is None:
        config = AgentGraphConfig()

    os.environ["LANGCHAIN_RECURSION_LIMIT"] = str(config.recursion_limit)

    agent_tools = list(tools)
    flow_control_tools: list[BaseTool] = create_flow_control_tools(response_format)
    llm_tools: list[BaseTool] = [*agent_tools, *flow_control_tools]

    init_node = create_init_node(messages)
    agent_node = create_llm_node(model, llm_tools)
    tool_nodes = create_tool_node(agent_tools)
    terminate_node = create_terminate_node(response_format)

    builder: StateGraph[AgentGraphState] = StateGraph(state_schema)
    builder.add_node(AgentGraphNode.INIT, init_node)
    builder.add_node(AgentGraphNode.AGENT, agent_node)

    for tool_name, tool_node in tool_nodes.items():
        builder.add_node(tool_name, tool_node)

    builder.add_node(AgentGraphNode.TERMINATE, terminate_node)

    builder.add_edge(START, AgentGraphNode.INIT)
    builder.add_edge(AgentGraphNode.INIT, AgentGraphNode.AGENT)

    tool_node_names = list(tool_nodes.keys())
    builder.add_conditional_edges(
        AgentGraphNode.AGENT,
        route_agent,
        [AgentGraphNode.AGENT, *tool_node_names, AgentGraphNode.TERMINATE],
    )

    for tool_name in tool_node_names:
        builder.add_edge(tool_name, AgentGraphNode.AGENT)

    builder.add_edge(AgentGraphNode.TERMINATE, END)

    return builder



import json
import os
import tempfile

import pytest

from uipath_langchain._cli.cli_run import langgraph_run_middleware


@pytest.fixture
def simple_agent() -> str:
    if os.path.isfile("mocks/simple_agent.py"):
        with open("mocks/simple_agent.py", "r") as file:
            data = file.read()
    else:
        with open("tests/cli/mocks/simple_agent.py", "r") as file:
            data = file.read()
    return data


@pytest.fixture
def uipath_json() -> str:
    if os.path.isfile("mocks/uipath.json"):
        with open("mocks/uipath.json", "r") as file:
            data = file.read()
    else:
        with open("tests/cli/mocks/uipath.json", "r") as file:
            data = file.read()
    return data


@pytest.fixture
def langgraph_json() -> str:
    if os.path.isfile("mocks/langgraph.json"):
        with open("mocks/langgraph.json", "r") as file:
            data = file.read()
    else:
        with open("tests/cli/mocks/langgraph.json", "r") as file:
            data = file.read()
    return data


class TestRun:
    def test_successful_execution(
        self,
        langgraph_json: str,
        uipath_json: str,
        simple_agent: str,
        mock_env_vars: dict[str, str],
    ):
        os.environ.clear()
        os.environ.update(mock_env_vars)
        input_file_name = "input.json"
        output_file_name = "output.json"
        agent_file_name = "main.py"
        input_json_content = {"topic": "UiPath"}
        with tempfile.TemporaryDirectory() as temp_dir:
            current_dir = os.getcwd()
            os.chdir(temp_dir)
            # Create input and output files
            input_file_path = os.path.join(temp_dir, input_file_name)
            output_file_path = os.path.join(temp_dir, output_file_name)

            with open(input_file_path, "w") as f:
                f.write(json.dumps(input_json_content))

            # Create test script
            script_file_path = os.path.join(temp_dir, agent_file_name)
            with open(script_file_path, "w") as f:
                f.write(simple_agent)

            # create uipath.json
            uipath_json_file_path = os.path.join(temp_dir, "uipath.json")
            with open(uipath_json_file_path, "w") as f:
                f.write(uipath_json)

            # Create langgraph.json
            langgraph_json_file_path = os.path.join(temp_dir, "langgraph.json")
            with open(langgraph_json_file_path, "w") as f:
                f.write(langgraph_json)

            result = langgraph_run_middleware(
                entrypoint="agent",
                input=None,
                resume=False,
                input_file=input_file_path,
                execution_output_file=output_file_path,
            )
            assert result.should_continue is False
            assert os.path.exists(output_file_path)
            with open(output_file_path, "r") as f:
                output = f.read()
                assert "This is mock report for" in output

            os.chdir(current_dir)



from enum import Enum
from typing import Optional, Union

from uipath._cli._runtime._contracts import (
    UiPathBaseRuntimeError,
    UiPathErrorCategory,
    UiPathErrorCode,
)


class LangGraphErrorCode(Enum):
    CONFIG_MISSING = "CONFIG_MISSING"
    CONFIG_INVALID = "CONFIG_INVALID"

    GRAPH_NOT_FOUND = "GRAPH_NOT_FOUND"
    GRAPH_IMPORT_ERROR = "GRAPH_IMPORT_ERROR"
    GRAPH_TYPE_ERROR = "GRAPH_TYPE_ERROR"
    GRAPH_VALUE_ERROR = "GRAPH_VALUE_ERROR"
    GRAPH_LOAD_ERROR = "GRAPH_LOAD_ERROR"
    GRAPH_INVALID_UPDATE = "GRAPH_INVALID_UPDATE"
    GRAPH_EMPTY_INPUT = "GRAPH_EMPTY_INPUT"

    DB_QUERY_FAILED = "DB_QUERY_FAILED"
    DB_TABLE_CREATION_FAILED = "DB_TABLE_CREATION_FAILED"
    HITL_EVENT_CREATION_FAILED = "HITL_EVENT_CREATION_FAILED"
    DB_INSERT_FAILED = "DB_INSERT_FAILED"
    LICENSE_NOT_AVAILABLE = "LICENSE_NOT_AVAILABLE"


class LangGraphRuntimeError(UiPathBaseRuntimeError):
    """Custom exception for LangGraph runtime errors with structured error information."""

    def __init__(
        self,
        code: Union[LangGraphErrorCode, UiPathErrorCode],
        title: str,
        detail: str,
        category: UiPathErrorCategory = UiPathErrorCategory.UNKNOWN,
        status: Optional[int] = None,
    ):
        super().__init__(
            code.value, title, detail, category, status, prefix="LANGGRAPH"
        )



"""Process tool creation for UiPath process execution."""

from __future__ import annotations

from typing import Any, Type

from jsonschema_pydantic import jsonschema_to_pydantic  # type: ignore[import-untyped]
from langchain_core.tools import StructuredTool
from langgraph.types import interrupt
from pydantic import BaseModel
from uipath.agent.models.agent import AgentProcessToolResourceConfig
from uipath.models import InvokeProcess

from .utils import sanitize_tool_name


def create_process_tool(resource: AgentProcessToolResourceConfig) -> StructuredTool:
    """Uses interrupt() to suspend graph execution until process completes (handled by runtime)."""
    tool_name: str = sanitize_tool_name(resource.name)
    process_name = resource.properties.process_name
    folder_path = resource.properties.folder_path

    input_model: Type[BaseModel] = jsonschema_to_pydantic(resource.input_schema)
    output_model: Type[BaseModel] = jsonschema_to_pydantic(resource.output_schema)

    async def process_tool_fn(**kwargs: Any):
        try:
            result = interrupt(
                InvokeProcess(
                    name=process_name,
                    input_arguments=kwargs,
                    process_folder_path=folder_path,
                    process_folder_key=None,
                )
            )
        except Exception:
            raise

        return result

    class ProcessTool(StructuredTool):
        """Process tool with OutputType for schema compatibility."""

        OutputType: Type[BaseModel] = output_model

    return ProcessTool(
        name=tool_name,
        description=resource.description,
        args_schema=input_model,
        coroutine=process_tool_fn,
    )



import os
import dotenv
import httpx
from contextlib import asynccontextmanager
from typing import Optional, Literal

from pydantic import BaseModel
from langgraph.graph import StateGraph, START, END
from langgraph.types import Command
from langgraph.prebuilt import create_react_agent
from langchain.schema import SystemMessage, HumanMessage

from uipath_langchain.chat.models import UiPathChat
from langchain_mcp_adapters.tools import load_mcp_tools
from mcp import ClientSession
from mcp.client.streamable_http import streamablehttp_client
from uipath import UiPath

dotenv.load_dotenv()

UIPATH_CLIENT_ID = "EXTERNAL_APP_CLIENT_ID_HERE"
UIPATH_CLIENT_SECRET = os.getenv("UIPATH_CLIENT_SECRET")
UIPATH_SCOPE = "OR.Jobs"
UIPATH_URL = "base_url"
UIPATH_MCP_SERVER_URL = os.getenv("UIPATH_MCP_SERVER_URL")

class GraphInput(BaseModel):
    task: str

class GraphOutput(BaseModel):
    result: str

class State(BaseModel):
    task: str
    access_token: Optional[str] = os.getenv("UIPATH_ACCESS_TOKEN")
    result: Optional[str] = None

async def fetch_new_access_token(state: State) -> Command:
    try:
        UiPath(
            base_url=UIPATH_URL,
            client_id=UIPATH_CLIENT_ID,
            client_secret=UIPATH_CLIENT_SECRET,
            scope=UIPATH_SCOPE,
        )
        return Command(update={"access_token": os.getenv("UIPATH_ACCESS_TOKEN")})

    except Exception as e:
        raise Exception(f"Failed to initialize UiPath SDK: {str(e)}")

@asynccontextmanager
async def agent_mcp(access_token: str):
    async with streamablehttp_client(
        url=UIPATH_MCP_SERVER_URL,
        headers={"Authorization": f"Bearer {access_token}"},
        timeout=60,
    ) as (read, write, _):
        async with ClientSession(read, write) as session:
            await session.initialize()
            tools = await load_mcp_tools(session)
            model = UiPathChat(model="anthropic.claude-3-5-sonnet-20240620-v1:0")
            agent = create_react_agent(model, tools=tools)
            yield agent

async def connect_to_mcp(state: State) -> Command:
    try:
        async with agent_mcp(state.access_token) as agent:
            agent_response = await agent.ainvoke({
                "messages": [
                    SystemMessage(content="You are a helpful assistant."),
                    HumanMessage(content=state.task),
                ],
            })
            return Command(update={"result": agent_response["messages"][-1].content})
    except ExceptionGroup as e:
        for error in e.exceptions:
            if isinstance(error, httpx.HTTPStatusError) and error.response.status_code == 401:
                return Command(update={"access_token": None})
        raise

def route_start(state: State) -> Literal["fetch_new_access_token", "connect_to_mcp"]:
    return "fetch_new_access_token" if state.access_token is None else "connect_to_mcp"

def route_after_connect(state: State):
    return "fetch_new_access_token" if state.access_token is None else END

builder = StateGraph(State, input=GraphInput, output=GraphOutput)
builder.add_node("fetch_new_access_token", fetch_new_access_token)
builder.add_node("connect_to_mcp", connect_to_mcp)

builder.add_conditional_edges(START, route_start)
builder.add_edge("fetch_new_access_token", "connect_to_mcp")
builder.add_conditional_edges("connect_to_mcp", route_after_connect)

graph = builder.compile()



"""Tool creation and management for LowCode agents."""

from .tool_factory import (
    create_tools_from_resources,
)
from .tool_node import create_tool_node

__all__ = ["create_tools_from_resources", "create_tool_node"]



from typing import Optional

from langgraph.checkpoint.sqlite.aio import AsyncSqliteSaver
from uipath._cli._runtime._contracts import UiPathRuntimeContext


class LangGraphRuntimeContext(UiPathRuntimeContext):
    """Context information passed throughout the runtime execution."""

    memory: Optional[AsyncSqliteSaver] = None



"""Tests for schema utility functions."""

from unittest.mock import MagicMock

from pydantic import BaseModel, Field

from uipath_langchain._cli._utils._schema import (
    generate_schema_from_graph,
    resolve_refs,
)


class TestResolveRefs:
    """Tests for the resolve_refs function."""

    def test_simple_schema_without_refs(self):
        """Should return schema unchanged when no $refs exist."""
        schema = {
            "type": "object",
            "properties": {"name": {"type": "string"}, "age": {"type": "integer"}},
        }

        result, has_circular = resolve_refs(schema)

        assert result == schema
        assert has_circular is False

    def test_simple_ref_resolution(self):
        """Should resolve a simple $ref to its definition."""
        schema = {
            "properties": {"user": {"$ref": "#/$defs/User"}},
            "$defs": {
                "User": {
                    "type": "object",
                    "properties": {"name": {"type": "string"}},
                }
            },
        }

        result, has_circular = resolve_refs(schema)

        assert result["properties"]["user"]["type"] == "object"
        assert result["properties"]["user"]["properties"]["name"]["type"] == "string"
        assert has_circular is False

    def test_circular_dependency_detection(self):
        """Should detect circular dependencies in schema."""
        schema = {
            "properties": {"node": {"$ref": "#/$defs/Node"}},
            "$defs": {
                "Node": {
                    "type": "object",
                    "properties": {
                        "value": {"type": "string"},
                        "next": {"$ref": "#/$defs/Node"},
                    },
                }
            },
        }

        result, has_circular = resolve_refs(schema)

        assert has_circular is True
        # Check that circular ref was replaced with simplified schema
        assert result["properties"]["node"]["properties"]["next"]["type"] == "object"
        assert (
            "Circular reference"
            in result["properties"]["node"]["properties"]["next"]["description"]
        )

    def test_nested_refs_in_properties(self):
        """Should resolve nested $refs in object properties."""
        schema = {
            "properties": {
                "person": {"$ref": "#/$defs/Person"},
                "address": {"$ref": "#/$defs/Address"},
            },
            "$defs": {
                "Person": {
                    "type": "object",
                    "properties": {"name": {"type": "string"}},
                },
                "Address": {
                    "type": "object",
                    "properties": {"street": {"type": "string"}},
                },
            },
        }

        result, has_circular = resolve_refs(schema)

        assert result["properties"]["person"]["type"] == "object"
        assert result["properties"]["person"]["properties"]["name"]["type"] == "string"
        assert result["properties"]["address"]["type"] == "object"
        assert (
            result["properties"]["address"]["properties"]["street"]["type"] == "string"
        )
        assert has_circular is False

    def test_refs_in_arrays(self):
        """Should resolve $refs inside array items."""
        schema = {
            "properties": {
                "users": {
                    "type": "array",
                    "items": {"$ref": "#/$defs/User"},
                }
            },
            "$defs": {
                "User": {
                    "type": "object",
                    "properties": {"id": {"type": "integer"}},
                }
            },
        }

        result, has_circular = resolve_refs(schema)

        assert result["properties"]["users"]["items"]["type"] == "object"
        assert (
            result["properties"]["users"]["items"]["properties"]["id"]["type"]
            == "integer"
        )
        assert has_circular is False

    def test_multiple_circular_dependencies(self):
        """Should handle multiple circular dependencies in same schema."""
        schema = {
            "properties": {
                "node1": {"$ref": "#/$defs/Node"},
                "node2": {"$ref": "#/$defs/Node"},
            },
            "$defs": {
                "Node": {
                    "type": "object",
                    "properties": {
                        "next": {"$ref": "#/$defs/Node"},
                    },
                }
            },
        }

        result, has_circular = resolve_refs(schema)

        assert has_circular is True


class TestGenerateSchemaFromGraph:
    """Tests for the generate_schema_from_graph function."""

    def test_graph_without_schemas(self):
        """Should return empty schemas when graph has no input/output schemas."""
        mock_graph = MagicMock()
        del mock_graph.input_schema
        del mock_graph.output_schema

        result = generate_schema_from_graph(mock_graph)

        assert result.schema["input"] == {
            "type": "object",
            "properties": {},
            "required": [],
        }
        assert result.schema["output"] == {
            "type": "object",
            "properties": {},
            "required": [],
        }
        assert result.has_input_circular_dependency is False
        assert result.has_output_circular_dependency is False

    def test_graph_with_simple_schemas(self):
        """Should extract input and output schemas from graph."""

        class InputModel(BaseModel):
            query: str = Field(description="User query")
            max_results: int = Field(default=10)

        class OutputModel(BaseModel):
            response: str = Field(description="Agent response")

        mock_graph = MagicMock()
        mock_graph.input_schema = InputModel
        mock_graph.output_schema = OutputModel

        result = generate_schema_from_graph(mock_graph)

        assert "query" in result.schema["input"]["properties"]
        assert "max_results" in result.schema["input"]["properties"]
        assert result.schema["input"]["properties"]["query"]["type"] == "string"
        assert "response" in result.schema["output"]["properties"]
        assert result.schema["output"]["properties"]["response"]["type"] == "string"
        assert result.has_input_circular_dependency is False
        assert result.has_output_circular_dependency is False

    def test_graph_with_circular_input_schema(self):
        """Should detect circular dependencies in input schema."""

        class NodeInput(BaseModel):
            value: str
            children: list["NodeInput"] = Field(default_factory=list)

        mock_graph = MagicMock()
        mock_graph.input_schema = NodeInput
        del mock_graph.output_schema

        result = generate_schema_from_graph(mock_graph)

        assert result.has_input_circular_dependency is True
        assert result.has_output_circular_dependency is False
        assert "value" in result.schema["input"]["properties"]

    def test_graph_with_circular_output_schema(self):
        """Should detect circular dependencies in output schema."""

        class TreeOutput(BaseModel):
            name: str
            parent: "TreeOutput | None" = None

        mock_graph = MagicMock()
        del mock_graph.input_schema
        mock_graph.output_schema = TreeOutput

        result = generate_schema_from_graph(mock_graph)

        assert result.has_input_circular_dependency is False
        assert result.has_output_circular_dependency is True
        assert "name" in result.schema["output"]["properties"]

    def test_graph_with_both_circular_schemas(self):
        """Should detect circular dependencies in both input and output schemas."""

        class CircularInput(BaseModel):
            data: str
            ref: "CircularInput | None" = None

        class CircularOutput(BaseModel):
            result: str
            next: "CircularOutput | None" = None

        mock_graph = MagicMock()
        mock_graph.input_schema = CircularInput
        mock_graph.output_schema = CircularOutput

        result = generate_schema_from_graph(mock_graph)

        assert result.has_input_circular_dependency is True
        assert result.has_output_circular_dependency is True
        assert "data" in result.schema["input"]["properties"]
        assert "result" in result.schema["output"]["properties"]

    def test_graph_with_required_fields(self):
        """Should extract required fields from schemas."""

        class StrictModel(BaseModel):
            required_field: str
            optional_field: str | None = None

        mock_graph = MagicMock()
        mock_graph.input_schema = StrictModel
        del mock_graph.output_schema

        result = generate_schema_from_graph(mock_graph)

        assert "required_field" in result.schema["input"]["required"]
        assert "optional_field" not in result.schema["input"]["required"]



import json
import os
import os.path
import shutil
import sqlite3
import uuid

from click.testing import CliRunner
from pytest_httpx import HTTPXMock

from tests.hitl.conftest import get_file_path
from uipath_langchain._cli.cli_run import langgraph_run_middleware


def get_org_scoped_url(base_url: str):
    return base_url.rsplit("/", 1)[0]


class TestHitlActionTrigger:
    """Test class for Action trigger functionality."""

    def test_agent_action_trigger(
        self,
        runner: CliRunner,
        temp_dir: str,
        httpx_mock: HTTPXMock,
        setup_test_env: None,
    ) -> None:
        script_name = "action_trigger_hitl.py"
        script_file_path = get_file_path(script_name)

        config_file_name = "uipath.json"
        config_file_path = get_file_path(config_file_name)

        langgraph_config_file_name = "langgraph.json"
        langgraph_config_file_path = get_file_path(langgraph_config_file_name)

        with runner.isolated_filesystem(temp_dir=temp_dir):
            # Copy the API trigger test file to our temp directory

            shutil.copy(script_file_path, "hitl.py")
            shutil.copy(config_file_path, config_file_name)
            shutil.copy(langgraph_config_file_path, "langgraph.json")

            # mock app creation
            base_url = os.getenv(
                "UIPATH_URL", "https://cloud.uipath.com/organization/tenant"
            )
            action_key = uuid.uuid4()

            # Mock UiPath API responses for action app creation
            httpx_mock.add_response(
                url=f"{base_url}/orchestrator_/tasks/AppTasks/CreateAppTask",
                json={
                    "id": 1,
                    "title": "Action Required: Report Review",
                    "key": f"{action_key}",
                },
            )

            httpx_mock.add_response(
                url=f"{get_org_scoped_url(base_url)}/apps_/default/api/v1/default/deployed-action-apps-schemas?search=HITL APP",
                text=json.dumps(
                    {
                        "deployed": [
                            {
                                "deploymentFolder": {
                                    "fullyQualifiedName": "app-folder-path"
                                },
                                "systemName": "HITL APP",
                                "actionSchema": {
                                    "key": "test-key",
                                    "inputs": [],
                                    "outputs": [],
                                    "inOuts": [],
                                    "outcomes": [],
                                },
                            }
                        ]
                    }
                ),
            )

            # First execution: creates action trigger and stores it in database
            result = langgraph_run_middleware("agent", "{}", False)

            assert result.error_message is None

            # Verify that __uipath directory and state.db were created
            assert os.path.exists("__uipath")
            assert os.path.exists("__uipath/state.db")

            # Verify the state database contains trigger information
            conn = None
            try:
                conn = sqlite3.connect("__uipath/state.db")
                cursor = conn.cursor()

                cursor.execute("""
                                        SELECT name FROM sqlite_master
                                        WHERE type='table' AND name='__uipath_resume_triggers'
                                    """)
                tables = cursor.fetchall()
                assert len(tables) == 1

                # Check the first action trigger data
                cursor.execute("SELECT * FROM __uipath_resume_triggers")
                triggers = cursor.fetchall()
                assert len(triggers) == 1
                _, type, key, folder_key, folder_path, payload, _ = triggers[0]
                assert type == "Task"
                assert folder_path == "app-folder-path"
                assert folder_key is None
                assert "agent question" in payload
                assert "Action Required" in payload
            finally:
                if conn:
                    conn.close()

            # Mock response for first resume: human response from create action
            httpx_mock.add_response(
                url=f"{base_url}/orchestrator_/tasks/GenericTasks/GetTaskDataByKey?taskKey={key}",
                json={
                    "id": 1,
                    "title": "Action Required: Report Review",
                    "data": {"Answer": "human response from create action"},
                },
            )

            # Second execution: resume from first trigger
            result = langgraph_run_middleware("agent", "{}", True)

            assert result.error_message is None
            assert result.should_continue is False

            # Verify second trigger information
            conn = None
            try:
                conn = sqlite3.connect("__uipath/state.db")
                cursor = conn.cursor()

                cursor.execute("""
                                                    SELECT name FROM sqlite_master
                                                    WHERE type='table' AND name='__uipath_resume_triggers'
                                                """)
                tables = cursor.fetchall()
                assert len(tables) == 1

                # Check the second trigger data (from wait action)
                cursor.execute("""SELECT * FROM __uipath_resume_triggers
                                  ORDER BY timestamp DESC
                                  """)
                triggers = cursor.fetchall()
                assert len(triggers) == 2
                _, type, key, folder_key, folder_path, payload, _ = triggers[0]
                assert type == "Task"
                assert folder_path is None
                assert folder_key is None
                assert "agent question from wait action" in payload
                assert key == "1662478a-65b4-4a09-8e22-d707e5bd64f3"
            finally:
                if conn:
                    conn.close()

            # Mock response for second resume: human response from wait action
            httpx_mock.add_response(
                url=f"{base_url}/orchestrator_/tasks/GenericTasks/GetTaskDataByKey?taskKey={key}",
                json={
                    "id": 1,
                    "title": "Action Required: Report Review",
                    "data": {"Answer": "human response from wait action"},
                },
            )

            # Third execution: resume from second trigger and complete
            result = langgraph_run_middleware("agent", "{}", True)
            assert result.error_message is None
            assert result.should_continue is False

            # Verify final output contains the last human response
            with open("__uipath/output.json", "r") as f:
                output = f.read()
            json_output = json.loads(output)
            assert json_output["output"] == {
                "message": "human response from wait action"
            }

            # Verify execution log contains both human responses
            with open("__uipath/execution.log", "r") as f:
                output = f.read()

            assert "Response from HITL action:" in output
            assert "human response from create action" in output



import logging
import os
from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator, AsyncIterator, Optional, Sequence
from uuid import uuid4

from langchain_core.runnables.config import RunnableConfig
from langgraph.checkpoint.sqlite.aio import AsyncSqliteSaver
from langgraph.errors import EmptyInputError, GraphRecursionError, InvalidUpdateError
from langgraph.graph.state import CompiledStateGraph, StateGraph
from langgraph.types import Interrupt, StateSnapshot
from typing_extensions import override
from uipath._cli._runtime._contracts import (
    UiPathBaseRuntime,
    UiPathBreakpointResult,
    UiPathErrorCategory,
    UiPathErrorCode,
    UiPathResumeTrigger,
    UiPathRuntimeResult,
    UiPathRuntimeStatus,
)
from uipath._cli.models.runtime_schema import Entrypoint
from uipath._events._events import (
    UiPathAgentMessageEvent,
    UiPathAgentStateEvent,
    UiPathRuntimeEvent,
)

from .._utils._schema import generate_schema_from_graph
from ._context import LangGraphRuntimeContext
from ._exception import LangGraphErrorCode, LangGraphRuntimeError
from ._graph_resolver import AsyncResolver, LangGraphJsonResolver
from ._input import get_graph_input
from ._output import create_and_save_resume_trigger, serialize_output

logger = logging.getLogger(__name__)


class LangGraphRuntime(UiPathBaseRuntime):
    """
    A runtime class implementing the async context manager protocol.
    This allows using the class with 'async with' statements.
    """

    def __init__(self, context: LangGraphRuntimeContext, graph_resolver: AsyncResolver):
        super().__init__(context)
        self.context: LangGraphRuntimeContext = context
        self.graph_resolver: AsyncResolver = graph_resolver
        self.resume_triggers_table: str = "__uipath_resume_triggers"

    @asynccontextmanager
    async def _get_or_create_memory(self) -> AsyncIterator[AsyncSqliteSaver]:
        """
        Get existing memory from context or create a new one.

        If memory is created, it will be automatically disposed at the end.
        If memory already exists in context, it will be reused without disposal.

        Yields:
            AsyncSqliteSaver instance
        """
        # Check if memory already exists in context
        if self.context.memory is not None:
            # Use existing memory, don't dispose
            yield self.context.memory
        else:
            # Create new memory and dispose at the end
            async with AsyncSqliteSaver.from_conn_string(
                self.state_file_path
            ) as memory:
                yield memory
                # Memory is automatically disposed by the context manager

    async def execute(self) -> Optional[UiPathRuntimeResult]:
        """Execute the graph with the provided input and configuration."""
        graph = await self.graph_resolver()
        if not graph:
            return None

        try:
            async with self._get_or_create_memory() as memory:
                compiled_graph = await self._setup_graph(memory, graph)
                graph_input = await self._get_graph_input(memory)
                graph_config = self._get_graph_config()

                # Execute without streaming
                graph_output = await compiled_graph.ainvoke(
                    graph_input,
                    graph_config,
                    interrupt_before=self.context.breakpoints,
                )

                # Get final state and create result
                self.context.result = await self._create_runtime_result(
                    compiled_graph, graph_config, memory, graph_output
                )

            return self.context.result

        except Exception as e:
            raise self._create_runtime_error(e) from e

    async def stream(
        self,
    ) -> AsyncGenerator[UiPathRuntimeEvent | UiPathRuntimeResult, None]:
        """
        Stream graph execution events in real-time.

        Yields UiPath UiPathRuntimeEvent instances (thin wrappers around framework data),
        then yields the final UiPathRuntimeResult as the last item.
        The result is also stored in self.context.result.

        Yields:
            - UiPathAgentMessageEvent: Wraps framework messages (BaseMessage, chunks, etc.)
            - UiPathAgentStateEvent: Wraps framework state updates
            - Final event: UiPathRuntimeResult or UiPathBreakpointResult

        Example:
            async for event in runtime.stream():
                if isinstance(event, UiPathRuntimeResult):
                    # Last event is the result
                    print(f"Final result: {event}")
                elif isinstance(event, UiPathAgentMessageEvent):
                    # Access framework-specific message
                    message = event.payload  # BaseMessage or AIMessageChunk
                    print(f"Message: {message.content}")
                elif isinstance(event, UiPathAgentStateEvent):
                    # Access framework-specific state
                    state = event.payload
                    print(f"Node {event.node_name} updated: {state}")

        Raises:
            LangGraphRuntimeError: If execution fails
        """
        graph = await self.graph_resolver()
        if not graph:
            return

        try:
            async with self._get_or_create_memory() as memory:
                compiled_graph = await self._setup_graph(memory, graph)
                graph_input = await self._get_graph_input(memory)
                graph_config = self._get_graph_config()

                # Track final chunk for result creation
                final_chunk: Optional[dict[Any, Any]] = None

                # Stream events from graph
                async for stream_chunk in compiled_graph.astream(
                    graph_input,
                    graph_config,
                    interrupt_before=self.context.breakpoints,
                    stream_mode=["messages", "updates"],
                    subgraphs=True,
                ):
                    _, chunk_type, data = stream_chunk

                    # Emit UiPathAgentMessageEvent for messages
                    if chunk_type == "messages":
                        if isinstance(data, tuple):
                            message, _ = data
                            event = UiPathAgentMessageEvent(
                                payload=message,
                                execution_id=self.context.execution_id,
                            )
                            yield event

                    # Emit UiPathAgentStateEvent for state updates
                    elif chunk_type == "updates":
                        if isinstance(data, dict):
                            final_chunk = data

                            # Emit state update event for each node
                            for node_name, agent_data in data.items():
                                if isinstance(agent_data, dict):
                                    state_event = UiPathAgentStateEvent(
                                        payload=agent_data,
                                        node_name=node_name,
                                        execution_id=self.context.execution_id,
                                    )
                                    yield state_event

                # Extract output from final chunk
                graph_output = self._extract_graph_result(
                    final_chunk, compiled_graph.output_channels
                )

                # Get final state and create result
                self.context.result = await self._create_runtime_result(
                    compiled_graph, graph_config, memory, graph_output
                )

                # Yield the final result as last event
                yield self.context.result

        except Exception as e:
            raise self._create_runtime_error(e) from e

    async def _setup_graph(
        self, memory: AsyncSqliteSaver, graph: StateGraph[Any, Any, Any]
    ) -> CompiledStateGraph[Any, Any, Any]:
        """Setup and compile the graph with memory and interrupts."""
        interrupt_before: list[str] = []
        interrupt_after: list[str] = []

        return graph.compile(
            checkpointer=memory,
            interrupt_before=interrupt_before,
            interrupt_after=interrupt_after,
        )

    def _get_graph_config(self) -> RunnableConfig:
        """Build graph execution configuration."""
        graph_config: RunnableConfig = {
            "configurable": {
                "thread_id": (
                    self.context.execution_id or self.context.job_id or "default"
                )
            },
            "callbacks": [],
        }

        # Add optional config from environment
        recursion_limit = os.environ.get("LANGCHAIN_RECURSION_LIMIT")
        max_concurrency = os.environ.get("LANGCHAIN_MAX_CONCURRENCY")

        if recursion_limit is not None:
            graph_config["recursion_limit"] = int(recursion_limit)
        if max_concurrency is not None:
            graph_config["max_concurrency"] = int(max_concurrency)

        return graph_config

    async def _get_graph_input(self, memory: AsyncSqliteSaver) -> Any:
        """Process and return graph input."""
        return await get_graph_input(
            context=self.context,
            memory=memory,
            resume_triggers_table=self.resume_triggers_table,
        )

    async def _get_graph_state(
        self,
        compiled_graph: CompiledStateGraph[Any, Any, Any],
        graph_config: RunnableConfig,
    ) -> Optional[StateSnapshot]:
        """Get final graph state."""
        try:
            return await compiled_graph.aget_state(graph_config)
        except Exception:
            return None

    def _extract_graph_result(
        self, final_chunk: Any, output_channels: str | Sequence[str]
    ) -> Any:
        """
        Extract the result from a LangGraph output chunk according to the graph's output channels.

        Args:
            final_chunk: The final chunk from graph.astream()
            output_channels: The graph's output channel configuration

        Returns:
            The extracted result according to the graph's output_channels configuration
        """
        # Unwrap from subgraph tuple format if needed
        if isinstance(final_chunk, tuple) and len(final_chunk) == 2:
            final_chunk = final_chunk[1]

        # If the result isn't a dict or graph doesn't define output channels, return as is
        if not isinstance(final_chunk, dict):
            return final_chunk

        # Case 1: Single output channel as string
        if isinstance(output_channels, str):
            return final_chunk.get(output_channels, final_chunk)

        # Case 2: Multiple output channels as sequence
        elif hasattr(output_channels, "__iter__") and not isinstance(
            output_channels, str
        ):
            # Check which channels are present
            available_channels = [ch for ch in output_channels if ch in final_chunk]

            # If no available channels, output may contain the last_node name as key
            unwrapped_final_chunk = {}
            if not available_channels and len(final_chunk) == 1:
                potential_unwrap = next(iter(final_chunk.values()))
                if isinstance(potential_unwrap, dict):
                    unwrapped_final_chunk = potential_unwrap
                    available_channels = [
                        ch for ch in output_channels if ch in unwrapped_final_chunk
                    ]

            if available_channels:
                # Create a dict with the available channels
                return {
                    channel: final_chunk.get(channel)
                    or unwrapped_final_chunk.get(channel)
                    for channel in available_channels
                }

        # Fallback for any other case
        return final_chunk

    def _is_interrupted(self, state: StateSnapshot) -> bool:
        """Check if execution was interrupted (static or dynamic)."""
        # Check for static interrupts (interrupt_before/after)
        if hasattr(state, "next") and state.next:
            return True

        # Check for dynamic interrupts (interrupt() inside node)
        if hasattr(state, "tasks"):
            for task in state.tasks:
                if hasattr(task, "interrupts") and task.interrupts:
                    return True

        return False

    def _get_dynamic_interrupt(self, state: StateSnapshot) -> Optional[Interrupt]:
        """Get the first dynamic interrupt if any."""
        if not hasattr(state, "tasks"):
            return None

        for task in state.tasks:
            if hasattr(task, "interrupts") and task.interrupts:
                for interrupt in task.interrupts:
                    if isinstance(interrupt, Interrupt):
                        return interrupt
        return None

    async def _create_runtime_result(
        self,
        compiled_graph: CompiledStateGraph[Any, Any, Any],
        graph_config: RunnableConfig,
        memory: AsyncSqliteSaver,
        graph_output: Optional[Any],
    ) -> UiPathRuntimeResult:
        """
        Get final graph state and create the execution result.

        Stores the result in self.context.result and self.context.state.

        Args:
            compiled_graph: The compiled graph instance
            graph_config: The graph execution configuration
            memory: The SQLite memory instance
            graph_output: The graph execution output
        """
        # Get the final state
        graph_state = await self._get_graph_state(compiled_graph, graph_config)

        # Check if execution was interrupted (static or dynamic)
        if graph_state and self._is_interrupted(graph_state):
            return await self._create_suspended_result(
                graph_state, memory, graph_output
            )
        else:
            # Normal completion
            return self._create_success_result(graph_output)

    async def _create_suspended_result(
        self,
        graph_state: StateSnapshot,
        graph_memory: AsyncSqliteSaver,
        graph_output: Optional[Any],
    ) -> UiPathRuntimeResult:
        """Create result for suspended execution."""
        # Check if it's a dynamic interrupt
        dynamic_interrupt = self._get_dynamic_interrupt(graph_state)
        resume_trigger: Optional[UiPathResumeTrigger] = None

        if dynamic_interrupt:
            # Dynamic interrupt - create and save resume trigger
            resume_trigger = await create_and_save_resume_trigger(
                interrupt_value=dynamic_interrupt.value,
                memory=graph_memory,
                resume_triggers_table=self.resume_triggers_table,
            )
            output = serialize_output(graph_output)
            return UiPathRuntimeResult(
                output=output,
                status=UiPathRuntimeStatus.SUSPENDED,
                resume=resume_trigger,
            )
        else:
            # Static interrupt (breakpoint)
            return self._create_breakpoint_result(graph_state)

    def _create_breakpoint_result(
        self,
        graph_state: StateSnapshot,
    ) -> UiPathBreakpointResult:
        """Create result for execution paused at a breakpoint."""

        # Get next nodes - these are the nodes that will execute when resumed
        next_nodes = list(graph_state.next)

        # Determine breakpoint type and node
        if next_nodes:
            # Breakpoint is BEFORE these nodes (interrupt_before)
            breakpoint_type = "before"
            breakpoint_node = next_nodes[0]
        else:
            # Breakpoint is AFTER the last executed node (interrupt_after)
            # Get the last executed node from tasks
            breakpoint_type = "after"
            if graph_state.tasks:
                # Tasks contain the nodes that just executed
                # Get the last task's name
                breakpoint_node = graph_state.tasks[-1].name
            else:
                # Fallback if no tasks (shouldn't happen)
                breakpoint_node = "unknown"

        return UiPathBreakpointResult(
            breakpoint_node=breakpoint_node,
            breakpoint_type=breakpoint_type,
            current_state=graph_state.values,
            next_nodes=next_nodes,
        )

    def _create_success_result(self, output: Optional[Any]) -> UiPathRuntimeResult:
        """Create result for successful completion."""
        return UiPathRuntimeResult(
            output=serialize_output(output),
            status=UiPathRuntimeStatus.SUCCESSFUL,
        )

    def _create_runtime_error(self, e: Exception) -> LangGraphRuntimeError:
        """Handle execution errors and create appropriate LangGraphRuntimeError."""
        if isinstance(e, LangGraphRuntimeError):
            return e

        detail = f"Error: {str(e)}"

        if isinstance(e, GraphRecursionError):
            return LangGraphRuntimeError(
                LangGraphErrorCode.GRAPH_LOAD_ERROR,
                "Graph recursion limit exceeded",
                detail,
                UiPathErrorCategory.USER,
            )

        if isinstance(e, InvalidUpdateError):
            return LangGraphRuntimeError(
                LangGraphErrorCode.GRAPH_INVALID_UPDATE,
                str(e),
                detail,
                UiPathErrorCategory.USER,
            )

        if isinstance(e, EmptyInputError):
            return LangGraphRuntimeError(
                LangGraphErrorCode.GRAPH_EMPTY_INPUT,
                "The input data is empty",
                detail,
                UiPathErrorCategory.USER,
            )

        return LangGraphRuntimeError(
            UiPathErrorCode.EXECUTION_ERROR,
            "Graph execution failed",
            detail,
            UiPathErrorCategory.USER,
        )

    async def validate(self) -> None:
        """Validate runtime inputs."""
        pass

    async def cleanup(self) -> None:
        """Cleanup runtime resources."""
        pass


class LangGraphScriptRuntime(LangGraphRuntime):
    """
    Resolves the graph from langgraph.json config file and passes it to the base runtime.
    """

    def __init__(
        self, context: LangGraphRuntimeContext, entrypoint: Optional[str] = None
    ):
        self.resolver = LangGraphJsonResolver(entrypoint=entrypoint)
        super().__init__(context, self.resolver)

    @override
    async def get_entrypoint(self) -> Entrypoint:
        """Get entrypoint for this LangGraph runtime."""
        graph = await self.resolver()
        compiled_graph = graph.compile()
        schema_details = generate_schema_from_graph(compiled_graph)

        return Entrypoint(
            file_path=self.context.entrypoint,  # type: ignore[call-arg]
            unique_id=str(uuid4()),
            type="agent",
            input=schema_details.schema["input"],
            output=schema_details.schema["output"],
        )

    async def cleanup(self) -> None:
        """Cleanup runtime resources including resolver."""
        await super().cleanup()
        await self.resolver.cleanup()



"""Exceptions for the basic agent loop."""

from uipath._cli._runtime._contracts import UiPathRuntimeError


class AgentNodeRoutingException(Exception):
    pass


class AgentTerminationException(UiPathRuntimeError):
    pass



import json
import os
import os.path
import shutil
import sqlite3
import uuid

from click.testing import CliRunner
from pytest_httpx import HTTPXMock

from tests.hitl.conftest import get_file_path
from uipath_langchain._cli.cli_run import langgraph_run_middleware


class TestHitlJobTrigger:
    """Test class for Job trigger functionality."""

    def test_agent_job_trigger(
        self,
        runner: CliRunner,
        temp_dir: str,
        httpx_mock: HTTPXMock,
        setup_test_env: None,
    ) -> None:
        script_name = "job_trigger_hitl.py"
        script_file_path = get_file_path(script_name)

        config_file_name = "uipath.json"
        config_file_path = get_file_path(config_file_name)

        langgraph_config_file_name = "langgraph.json"
        langgraph_config_file_path = get_file_path(langgraph_config_file_name)

        with runner.isolated_filesystem(temp_dir=temp_dir):
            # Copy the API trigger test file to our temp directory

            shutil.copy(script_file_path, "hitl.py")
            shutil.copy(config_file_path, config_file_name)
            shutil.copy(langgraph_config_file_path, "langgraph.json")

            # mock app creation
            base_url = os.getenv("UIPATH_URL")
            job_key = uuid.uuid4()

            # Mock UiPath API response for job creation
            httpx_mock.add_response(
                url=f"{base_url}/orchestrator_/odata/Jobs/UiPath.Server.Configuration.OData.StartJobs",
                json={"value": [{"key": f"{job_key}", "Id": "123"}]},
            )

            # First execution: creates job trigger and stores it in database
            result = langgraph_run_middleware("agent", "{}", False)

            assert result.error_message is None

            # Verify that __uipath directory and state.db were created
            assert os.path.exists("__uipath")
            assert os.path.exists("__uipath/state.db")

            # Verify the state database contains trigger information
            conn = None
            try:
                conn = sqlite3.connect("__uipath/state.db")
                cursor = conn.cursor()

                cursor.execute("""
                                        SELECT name FROM sqlite_master
                                        WHERE type='table' AND name='__uipath_resume_triggers'
                                    """)
                tables = cursor.fetchall()
                assert len(tables) == 1

                # Check the first job trigger data
                cursor.execute("SELECT * FROM __uipath_resume_triggers")
                triggers = cursor.fetchall()
                assert len(triggers) == 1
                _, type, key, folder_key, folder_path, payload, _ = triggers[0]
                assert type == "Job"
                assert folder_path == "process-folder-path"
                assert folder_key is None
                assert "input_arg_1" in payload
                assert "value_1" in payload
            finally:
                if conn:
                    conn.close()

            # Mock response for first resume: job output arguments
            output_args_dict = {"output_arg_1": "response from invoke process"}
            httpx_mock.add_response(
                url=f"{base_url}/orchestrator_/odata/Jobs/UiPath.Server.Configuration.OData.GetByKey(identifier={key})",
                json={
                    "key": f"{job_key}",
                    "id": 123,
                    "output_arguments": json.dumps(output_args_dict),
                },
            )
            # Second execution: resume from first trigger
            result = langgraph_run_middleware("agent", "{}", True)
            assert result.error_message is None
            assert result.should_continue is False

            # Verify second trigger information
            conn = None
            try:
                conn = sqlite3.connect("__uipath/state.db")
                cursor = conn.cursor()

                cursor.execute("""
                                                    SELECT name FROM sqlite_master
                                                    WHERE type='table' AND name='__uipath_resume_triggers'
                                                """)
                tables = cursor.fetchall()
                assert len(tables) == 1

                # Check the second job trigger data (from wait job)
                cursor.execute("""SELECT * FROM __uipath_resume_triggers
                                  ORDER BY timestamp DESC
                                  """)
                triggers = cursor.fetchall()
                assert len(triggers) == 2
                _, type, key, folder_key, folder_path, payload, _ = triggers[0]
                assert type == "Job"
                assert folder_path is None
                assert folder_key is None
                assert "123" in payload
                assert key == "487d9dc7-30fe-4926-b5f0-35a956914042"
            finally:
                if conn:
                    conn.close()

            # Mock response for second resume: wait job output arguments
            output_args_dict = {"output_arg_2": "response from wait job"}

            httpx_mock.add_response(
                url=f"{base_url}/orchestrator_/odata/Jobs/UiPath.Server.Configuration.OData.GetByKey(identifier={key})",
                json={
                    "key": f"{job_key}",
                    "id": 123,
                    "output_arguments": json.dumps(output_args_dict),
                },
            )

            # Third execution: resume from second trigger and complete
            result = langgraph_run_middleware("agent", "{}", True)
            assert result.error_message is None
            assert result.should_continue is False

            # Verify final output contains the last job response
            with open("__uipath/output.json", "r") as f:
                output = f.read()
            json_output = json.loads(output)
            assert json_output["output"] == {"message": "response from wait job"}

            # Verify execution log contains both job responses
            with open("__uipath/execution.log", "r") as f:
                output = f.read()

            assert "Process output" in output
            assert "response from invoke process" in output



"""Runtime factory for LangGraph projects."""

from uipath._cli._runtime._contracts import UiPathRuntimeFactory

from ._cli._runtime._context import LangGraphRuntimeContext
from ._cli._runtime._runtime import LangGraphScriptRuntime


class LangGraphRuntimeFactory(
    UiPathRuntimeFactory[LangGraphScriptRuntime, LangGraphRuntimeContext]
):
    """Factory for LangGraph runtimes."""

    def __init__(self):
        super().__init__(
            LangGraphScriptRuntime,
            LangGraphRuntimeContext,
            context_generator=lambda **kwargs: LangGraphRuntimeContext.with_defaults(
                **kwargs
            ),
        )



import asyncio
from typing import List, Optional

from openinference.instrumentation.langchain import (
    LangChainInstrumentor,
    get_current_span,
)
from uipath._cli._evals._console_progress_reporter import ConsoleProgressReporter
from uipath._cli._evals._evaluate import evaluate
from uipath._cli._evals._progress_reporter import StudioWebProgressReporter
from uipath._cli._evals._runtime import UiPathEvalContext
from uipath._cli._runtime._contracts import (
    UiPathRuntimeFactory,
)
from uipath._cli._utils._eval_set import EvalHelpers
from uipath._cli.middlewares import MiddlewareResult
from uipath._events._event_bus import EventBus
from uipath.eval._helpers import auto_discover_entrypoint
from uipath.tracing import LlmOpsHttpExporter

from uipath_langchain._cli._runtime._context import LangGraphRuntimeContext
from uipath_langchain._cli._runtime._runtime import LangGraphScriptRuntime
from uipath_langchain._cli._utils._graph import LangGraphConfig
from uipath_langchain._tracing import (
    _instrument_traceable_attributes,
)


def langgraph_eval_middleware(
    entrypoint: Optional[str], eval_set: Optional[str], eval_ids: List[str], **kwargs
) -> MiddlewareResult:
    config = LangGraphConfig()
    if not config.exists:
        return MiddlewareResult(
            should_continue=True
        )  # Continue with normal flow if no langgraph.json

    try:
        _instrument_traceable_attributes()

        event_bus = EventBus()

        if kwargs.get("register_progress_reporter", False):
            progress_reporter = StudioWebProgressReporter(
                spans_exporter=LlmOpsHttpExporter(extra_process_spans=True)
            )
            asyncio.run(progress_reporter.subscribe_to_eval_runtime_events(event_bus))
        console_reporter = ConsoleProgressReporter()
        asyncio.run(console_reporter.subscribe_to_eval_runtime_events(event_bus))

        def generate_runtime_context(
            context_entrypoint: str, **context_kwargs
        ) -> LangGraphRuntimeContext:
            context = LangGraphRuntimeContext.with_defaults(**context_kwargs)
            context.entrypoint = context_entrypoint
            return context

        runtime_entrypoint = entrypoint or auto_discover_entrypoint()

        eval_context = UiPathEvalContext.with_defaults(
            entrypoint=runtime_entrypoint, **kwargs
        )
        eval_context.eval_set = eval_set or EvalHelpers.auto_discover_eval_set()
        eval_context.eval_ids = eval_ids

        def generate_runtime(ctx: LangGraphRuntimeContext) -> LangGraphScriptRuntime:
            return LangGraphScriptRuntime(ctx, ctx.entrypoint)

        runtime_factory = UiPathRuntimeFactory(
            LangGraphScriptRuntime,
            LangGraphRuntimeContext,
            context_generator=lambda **context_kwargs: generate_runtime_context(
                context_entrypoint=runtime_entrypoint,
                **context_kwargs,
            ),
            runtime_generator=generate_runtime,
        )

        if eval_context.job_id:
            runtime_factory.add_span_exporter(
                LlmOpsHttpExporter(extra_process_spans=True)
            )

        runtime_factory.add_instrumentor(LangChainInstrumentor, get_current_span)

        asyncio.run(evaluate(runtime_factory, eval_context, event_bus))
        return MiddlewareResult(should_continue=False)

    except Exception as e:
        return MiddlewareResult(
            should_continue=False, error_message=f"Error running evaluation: {str(e)}"
        )



import json
import os
import shutil
import sqlite3

from click.testing import CliRunner
from pytest_httpx import HTTPXMock

from tests.hitl.conftest import get_file_path
from uipath_langchain._cli.cli_run import langgraph_run_middleware


class TestHitlApiTrigger:
    """Test class for HITL API trigger functionality."""

    def test_agent(
        self,
        runner: CliRunner,
        temp_dir: str,
        httpx_mock: HTTPXMock,
        setup_test_env: None,
    ) -> None:
        script_name = "api_trigger_hitl.py"
        script_file_path = get_file_path(script_name)

        config_file_name = "uipath.json"
        config_file_path = get_file_path(config_file_name)

        langgraph_config_file_name = "langgraph.json"
        langgraph_config_file_path = get_file_path(langgraph_config_file_name)

        with runner.isolated_filesystem(temp_dir=temp_dir):
            # Copy the API trigger test file to our temp directory

            shutil.copy(script_file_path, "hitl.py")
            shutil.copy(config_file_path, config_file_name)
            shutil.copy(langgraph_config_file_path, langgraph_config_file_name)

            # First execution: creates interrupt and stores trigger in database
            result = langgraph_run_middleware("agent", "{}", False)

            assert result.error_message is None

            # Verify that __uipath directory and state.db were created
            assert os.path.exists("__uipath")
            assert os.path.exists("__uipath/state.db")

            # Verify the state database contains trigger information
            conn = None
            try:
                conn = sqlite3.connect("__uipath/state.db")
                cursor = conn.cursor()

                cursor.execute("""
                    SELECT name FROM sqlite_master
                    WHERE type='table' AND name='__uipath_resume_triggers'
                """)
                tables = cursor.fetchall()
                assert len(tables) == 1

                # Check the inserted trigger data from first execution
                cursor.execute("SELECT * FROM __uipath_resume_triggers")
                triggers = cursor.fetchall()
                assert len(triggers) == 1
                _, type, key, folder_path, folder_key, payload, _ = triggers[0]
                assert type == "Api"
                assert folder_path == folder_key is None
                assert payload == "interrupt message"
            finally:
                if conn:
                    conn.close()

            # Mock API response for resume scenario
            base_url = os.getenv("UIPATH_URL")
            httpx_mock.add_response(
                url=f"{base_url}/orchestrator_/api/JobTriggers/GetPayload/{key}",
                status_code=200,
                text=json.dumps({"payload": "human response"}),
            )
            # Second execution: resume from stored trigger and fetch human response
            result = langgraph_run_middleware("agent", "{}", True)
            assert result.error_message is None
            assert result.should_continue is False

            # Verify the final output contains the resumed data
            with open("__uipath/output.json", "r") as f:
                output = f.read()
            json_output = json.loads(output)
            assert json_output["output"] == {"message": "human response"}



from __future__ import annotations

from enum import StrEnum

from langgraph.graph import MessagesState
from pydantic import BaseModel, Field


class AgentGraphState(MessagesState):
    """Agent Graph state for standard loop execution."""

    pass


class AgentGraphNode(StrEnum):
    INIT = "init"
    AGENT = "agent"
    TOOLS = "tools"
    TERMINATE = "terminate"


class AgentGraphConfig(BaseModel):
    recursion_limit: int = Field(
        default=50, ge=1, description="Maximum recursion limit for the agent graph"
    )
